/*
======================================================================
ode.cpp

This is the "Lightwave ODE Project"

Kevin MacPhail  07/05/04

	Updated		09/14/06


=======================================================
Some additional matrix wisdom from the ODE mailing list
=======================================================

Functions in ODE to multiply vectors by matrices, the standard thing to do
is to rotate a vector by a matrix.

dMultiply0, dMultiply1, dMultiply2.

To rotate a vector V by a matrix M, try:

dMultiply0(Result,M,V,3,3,1);

To multiply matrices together use instead:

dMultiply0(Result,M1,M2,3,3,3);

If you want to do it while at the same time transposing the matrix, use 
dMultiply1, dMultiply2 transposes both arguments (I believe, don't 
remember well).

--

How to determine the position and orientation of the objects.

You can get position (vector) and orientation (quaternion or 3x3 matrix) of
the bodies/geoms (that's described in user's guide, section 6.2) - and use
that for rendering.

4x4 matrix is just a composition of 3x3 rotation matrix, position vector and
a column/row containing (0,0,0,1). Depending on the graphics API you use
(opengl/d3d), matrices are laid out in column-major or row-major order.

For OpenGL, a 4x4 transformation matrix is:

Xx Yx Zx Ox
Xy Yy Zy Oy
Xz Yz Zz Oz
0  0  0  1

Here, upper-left 3x3 part is a rotation matrix (with X,Y,Z being axis
vectors, and x,y,z being their components) - you get that directly from ODE,
or convert quaternion into it. The Ox,Oy,Oz part is the position vector.

For D3D, the matrix is transposed (flipped along main diagonal):

Xx Xy Xz 0
Yx Yy Yz 0
Zx Zy Zz 0
Ox Oy Oz 1

Thus you have to transpose the rotation matrix obtained from ODE as well
(ODE uses OpenGL-style matrices)

--

ODE uses a +Z up coordinate system


       Z
       |
       |
       |
       |______Y
      /
    /
  X


Lightwave uses a +Y up coordinate system


       Y
       |
       |
       |
       |______Z
      /
    /
  X

So, to convert a +Z up coordinate system to a +Y up coordinate system
we need to swap the axis, i.e. swap both the Z & Y rows and columns, then 
convert from row-major to column-major.

0 1  2    0 1  2    0  2 1    0  8 4
4 5  6 -> 8 9 10 -> 8 10 9 -> 2 10 6
8 9 10    4 5  6    4  6 5    1  9 5

So, the order to read in the ODE 3x3 rotation matrix is: 0 8 4 2 10 6 1 9 5

====================================================================== */



extern "C" {
#include <lwserver.h>			// all plug-ins need this       
#include <lwgeneric.h>			// for the LayoutGeneric class  
#include <lwpanel.h>			// for "classic" panels
#include <lwhost.h>				// for the LWMessageFuncs global 
#include <lwrender.h>			// for the LWRenderFuncs global  
#include <lwcomlib.h>			// for the common library	     
#include <com_math.h>
#include <com_vecmatquat.h>
#include <lwhandler.h>
#include <lwmeshes.h>
#include <lwdisplce.h>
#include <lwenvel.h>
#include <lwmath.h>
}

#include <stdio.h>				// for NULL #define
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ode/ode.h>


#ifdef _MSC_VER
#pragma warning(disable:4244 4305)  // for VC++, no precision loss complaints
#endif


/////////////////////////////////////////////////
// debug
/////////////////////////////////////////////////
#define DEBUG 0


////////////////////////////////////////////////////////////////////
// some ode constants
// these may change
////////////////////////////////////////////////////////////////////
#define NUM 100			// max number of objects
#define MAXC 10			// max number of children
#define DENSITY (5.0)	// density of all objects
#define GPB 3			// maximum number of geometries per body
#define MAX_CONTACTS 8	// maximum number of contact points per body (was 4)
#define SPHERE 1
#define BOX 2
#define CYLINDER 3		// not implemented yet
#define TRIMESH 4		// not implemented yet
#define NONSTATIC 0
#define STATIC 1



/////////////////////////////////////////////////
// Displacment DEFINEs
/////////////////////////////////////////////////
#define INF_PIVLOCAL 0
#define ODEDISPLACE_VERSION 1


//////////////////////////////////
// Someday these may be translated
////////////////////////////////// 
#define STR_Type_TEXT      "Object Type"


///////////////////////////////
// Bounding Box Data Structure
///////////////////////////////
typedef struct st_BBox {
	LWDVector bbox[11];
	LWItemID id;
} BBox;


///////////////////////////
// The objMatrixData struct
///////////////////////////
typedef struct MyObject {
	dBodyID body;
	dGeomID geom[GPB];
} MyObject;


////////////////////////////////////////////////////////////////////////////////////////////
// For the ObjectData data structure
////////////////////////////////////////////////////////////////////////////////////////////
typedef struct st_ObjectData{
	LWDMatrix4* m;				//Lightwave rotation matrix
	dMatrix3 rotationMatrix;	//ODE rotation matrix
	LWItemID item;
	LWItemID parent;
	LWItemID rootParent;
	LWDVector objDelta;			// silly vector, try to remove the need for this
	dGeomID cg[MAXC];			//child geometry
	int objType;
	int objStatic;
	int numChildren;
	int odeNum;
	float xpos, ypos, zpos;		// the object's starting position
	struct st_ObjectData *next;
	BBox* bounds;				// contains bounds and position/rotation data
	float radius;				// radius of a sphere
} ObjectData;


//////////////////////////////
// Displacement Instance data/
/////////////////////////////
typedef struct st_ODEDisplaceData {
   LWTime		time;			// current time
   LWFrame		frame;			// current frame
   LWItemID		item;
   int			flags;			// use INF_ flag bit definitions
   int			objType;
   int			objStatic;
   char        desc[ 80 ];		// Your Message Here!
} ODEDisplaceData;


//////////////
// Prototypes
//////////////
XCALL_( int ) ODEGeneric( long version, GlobalFunc *global, LWLayoutGeneric *local, void *serverData );
void ODE_event( LWControl *ctrl, LWLayoutGeneric *local);
static void nearCallback (void *data, dGeomID o1, dGeomID o2);
XCALL_( static LWError )AddObject(LWItemID id);
void RemoveObject(LWItemID id);
ObjectData* FindObject( LWItemID id);
int CalcAB (LWItemID id_a, LWItemID id_b, LWDVector delta );
//int CalcRotAB (LWItemID id_a, LWItemID id_b, LWDVector delta );
int CalcBB ( LWItemID id );
static int pntScanBB( ObjectData* test, LWPntID id );
LWItemID findRootParents(ObjectData* test);
int sortList();
void KM_transpose4 ( LWDMatrix4 n, LWDMatrix4* m, int i );
void KM_transpose34 ( dMatrix3 n, dMatrix3 m );
void KM_dmatmul34( dMatrix3 a, dMatrix3 b, dMatrix3 c );
void KM_dtransform( LWDVector a, dMatrix3 m, LWDVector b );
void KM_dinitm34( dMatrix3 m, dReal a1, dReal b1, dReal c1, dReal d1,
							  dReal a2, dReal b2, dReal c2, dReal d2,
							  dReal a3, dReal b3, dReal c3, dReal d3 );


///////////////////
//For ODE_Displace
///////////////////
XCALL_( static int )popCnt_VMAP( void *data );
XCALL_( static const char * )popName_VMAP( void *data, int idx );
XCALL_( static int )popCnt_Item( void *data );
XCALL_( static const char * )popName_Item( void *data, int idx );
static LWItemID popItemID( int idx );
XCALL_( static LWInstance )Create( void *priv, LWItemID item, LWError *err );
XCALL_( static void )Destroy( ODEDisplaceData *dat );
XCALL_( static LWError )Copy( ODEDisplaceData *to, ODEDisplaceData *from );
XCALL_( static LWError )Load( ODEDisplaceData *dat, const LWLoadState *ls );
XCALL_( static LWError )Save( ODEDisplaceData *dat, const LWSaveState *ss );
XCALL_( static const char * )Describe( ODEDisplaceData *dat );
XCALL_( static const LWItemID * )UseItems( ODEDisplaceData *dat );
XCALL_( static void )ChangeID( ODEDisplaceData *dat, const LWItemID *ids );
XCALL_( static LWError )Init( ODEDisplaceData *dat, int mode );
XCALL_( static void )Cleanup( ODEDisplaceData *dat );
XCALL_( static LWError )NewTime( ODEDisplaceData *dat, LWFrame fr, LWTime t );
XCALL_( static int )Flags( ODEDisplaceData *dat );
XCALL_( static void )Evaluate( ODEDisplaceData *dat, LWDisplacementAccess *da );
XCALL_( int )ODEDisplace( long version, GlobalFunc *global, LWDisplacementHandler *local, void *serverData);
ODEDisplace_UI( long version, GlobalFunc *global, LWInterface *local, void *serverData );


/////////////////////////////
// ODE globals
// Need to consolidate style
/////////////////////////////
LWMessageFuncs *msg;			/* the message functions         */
LWPanelFuncs *panf;				/* the panel functions           */
LWSceneInfo *lwsi;				/* the scene functions           */
LWInterfaceInfo *intinfo;		/* the interface functions       */
LWPanelID panel;				/* the panel                     */
LWPanControlDesc desc;			/* used by macros in lwpanel.h   */
static LWObjectFuncs   *objf     = NULL;
static LWEnvelopeFuncs *envf     = NULL;
static LWChannelInfo   *chanf    = NULL;
static LWItemInfo      *iteminfo = NULL;
static LWObjectInfo    *objinfo;
static LWValue
   ival    = { LWT_INTEGER },
   ivecval = { LWT_VINT    },
   fval    = { LWT_FLOAT   },
   fvecval = { LWT_VFLOAT  },
   sval    = { LWT_STRING  };
static LWInstUpdate    *lwupdate = NULL;
static LWXPanelFuncs   *xpanf    = NULL;


//////////////////////////////////
// XPanels Stuff for ODE_Displace
//////////////////////////////////
enum { CH_TYPE = 0x8001, CH_STATIC };

static LWXPanelControl ctrl_list[] = {
   { CH_TYPE, STR_Type_TEXT,  "iChoice" },
   { CH_STATIC, "Static",  "iBoolean" },
   { 0 }
};

static LWXPanelDataDesc data_descrip[] = {
   { CH_TYPE, STR_Type_TEXT,  "integer" },
   { CH_STATIC, "Static",  "integer" },
   { 0 }
};

const char *objType_strlist[] = {
  "SPHERE",
  "BOX",
  NULL
};


//////////////////////////////////////////////////////////////////////////
//A global pointer to the array to hold the per-frame translation data for
//a single object. Will be read by the Displacement plugin included
//in this file.
//////////////////////////////////////////////////////////////////////////
ObjectData* objMatrixData = NULL;


//////////////////////////////////
// ODE Global
// dynamics and collision objects
//////////////////////////////////
static int num=0;					// number of objects in simulation
static int nextobj=0;				// next object to recycle if num==NUM
static dWorldID world;
static dSpaceID space;
static MyObject obj[NUM];			// Max Number of possible objects
static dJointGroupID contactgroup;
static int selected = -1;			// selected object
static int show_aabb = 0;			// show geom AABBs?
static int show_contacts = 0;		// show contact points?
static int random_pos = 1;			// drop objects from random position?
static int write_world = 0;


/*
======================================================================
ODEGeneric()

The ODE activation function.
====================================================================== */

XCALL_( int )
ODEGeneric( long version, GlobalFunc *global, LWLayoutGeneric *local, void *serverData )
{
   int ok;
   LWControl *ctl[10];


   if ( version != LWLAYOUTGENERIC_VERSION )
      return AFUNC_BADVERSION;


   ////////////////////////////////////////
   // get the message and panels functions
   ////////////////////////////////////////

   msg = (struct st_LWMessageFuncs *)global( LWMESSAGEFUNCS_GLOBAL, GFUSE_TRANSIENT );
   panf = (struct st_LWPanelFuncs *)global( LWPANELFUNCS_GLOBAL, GFUSE_TRANSIENT );
   lwsi = (struct st_LWSceneInfo *)global( LWSCENEINFO_GLOBAL, GFUSE_TRANSIENT );
   intinfo = (struct st_LWInterfaceInfo *)global( LWINTERFACEINFO_GLOBAL, GFUSE_TRANSIENT );
   objinfo  = (struct st_LWObjectInfo *)global( LWOBJECTINFO_GLOBAL, GFUSE_TRANSIENT );
   if ( !msg || !panf || !lwsi || !intinfo || !objinfo )
      return AFUNC_BADGLOBAL;


   //////////////////////////////////////////////////////
   // initialize the panels functions and create a panel
   //////////////////////////////////////////////////////

   panf->globalFun = global;

   panel = PAN_CREATE( panf, "LightWave ODE Project v0.2.0 BETA" );
   if ( !panel ) return AFUNC_BADGLOBAL;


   /////////////////////////////////////////////////////////////////
   // create an edit field on the panel and initialize its contents
   /////////////////////////////////////////////////////////////////
   ctl[0] = BUTTON_CTL( panf, panel, "Calculate This!" );
   CON_SETEVENT( ctl[0], ODE_event, local);


   //////////////////////////////////////////////////////////
   // display the panel; this waits until the user closes it
   //////////////////////////////////////////////////////////
   ok = panf->open( panel, PANF_BLOCKING | PANF_CANCEL );


   //////////////////////////////////////////////////////////////
   // ok is TRUE if the user pressed OK, FALSE if Cancel pressed
   //////////////////////////////////////////////////////////////
   //if ( ok )
      //GET_STR( ctl[1], edit, sizeof( edit ));
	

   //////////////////
   // free the panel
   //////////////////
   PAN_KILL( panf, panel );


   /////////
   // done!
   /////////
   return AFUNC_OK;

}


/*
======================================================================
ODE_event()

Button callback.

This is the callback that is activated when the "Calculate" button is
pushed.

====================================================================== */

void ODE_event( LWControl *ctrl, LWLayoutGeneric *local)
{
	////////////////////////////////////////////
	// Start the ODE calculations
	// This is where the heavy lifing gets done
	////////////////////////////////////////////
	LWDVector objPos;
	LWDVector objDelta;
	LWDVector objDeltaRotated;
	dMatrix3 invRotMatrix;
	dMatrix3 RM;
	ObjectData* test = objMatrixData;
	ObjectData* rp;
	ObjectData* rparent;
	dMass mass;
	int i; // current frame
	int frameStart, frameEnd, framesPerSecond;
	int totalFrames;
	int num = 0;
	int numC = 0;
	float stepRate;
	float lx, ly, lz;
	char buf[80];
	const dReal *R;		//pointer to a 4X4 matrix
	const dReal *pos;

	//////////////////////////////////////////////////////////////////
	//Get the scene frame info and calculate stepRate and totalFrames
	//////////////////////////////////////////////////////////////////
	frameStart = intinfo->previewStart;
	frameEnd = intinfo->previewEnd;
	framesPerSecond = lwsi->framesPerSecond;
	stepRate = 1.0/framesPerSecond;
	totalFrames = (frameEnd-frameStart)+1;


	//////////////////////////////////////////////
	// no objects added to the simulation so quit
	//////////////////////////////////////////////
	if (test == NULL) return;


	/////////////////////////////
	// Some Debugging statements
	/////////////////////////////
	if (DEBUG) {
		sprintf (buf, "The frameStart is %d", frameStart);
		msg->info( buf, NULL );
		sprintf(buf, "The stepRate is %f", stepRate);
		msg->info( buf, NULL );
		sprintf(buf, "The totalFrames is %d", totalFrames);
		msg->info( buf, NULL );
	}


	//////////////////////////////////////////////////////////////////
	// Reset the starting position of each object and re-evaluate the
	// objects for changed parenting data for each re-calculation
	//////////////////////////////////////////////////////////////////
	while (test != NULL)
	{
		iteminfo->param( test->item, LWIP_W_POSITION, 0, objPos );

		test->xpos = objPos[0];
		test->ypos = objPos[1];
		test->zpos = objPos[2];

		test->parent = iteminfo->parent(test->item);

		test = test->next;
	}

	
	sprintf (buf, "GoToFrame %d", frameStart);
	local->evaluate( local->data, buf );


	////////////////////////////////////////////////
	// Create the ODE world
	////////////////////////////////////////////////

	world = dWorldCreate();
	space = dHashSpaceCreate (0);
	contactgroup = dJointGroupCreate (0);
	dWorldSetGravity (world,0,0,-2.0);
	dWorldSetCFM (world,1e-5);
	dWorldSetAutoDisableFlag (world,1);
	dWorldSetContactMaxCorrectingVel (world,0.1);
	dWorldSetContactSurfaceLayer (world,0.001);
	dCreatePlane (space,0,0,1,0); // This could be set by the user later
	memset (obj,0,sizeof(obj));

	
	////////////////////////////////////////////////////////////////////////
	// Find the root parent of each object in the objMatrixData Linked List,
	// if any.
	////////////////////////////////////////////////////////////////////////

	test = objMatrixData; // a pointer to the linked list of objects

	while (test != NULL)
	{
		if (test->rootParent == NULL) //test only if root parent unknown
		{
			test->rootParent = findRootParents(test);		
		}
		test = test->next;
	}	

	sortList(); //guarantee rootParents are added to ODE first

	////////////////////////////////////////////////////////////////////////
	// Add a Body and Geom for each object in the objMatrixData Linked List
	////////////////////////////////////////////////////////////////////////

	test = objMatrixData; // reset the pointer to the linked list

	while (test != NULL)
	{
		if ( CalcBB( test->item ))
			return; //error

		//sprintf(buf, "%f ", test->radius);
		//msg->info( buf, NULL );

		lx = test->bounds->bbox[1][0] - test->bounds->bbox[0][0];
		ly = test->bounds->bbox[1][2] - test->bounds->bbox[0][2];
		lz = test->bounds->bbox[1][1] - test->bounds->bbox[0][1];

		// Matrix to rotate from world to item coordinates
		test->rotationMatrix[0] = test->bounds->bbox[5][0];
		test->rotationMatrix[1] = test->bounds->bbox[5][2];
		test->rotationMatrix[2] = test->bounds->bbox[5][1];
		test->rotationMatrix[3] = 0;
		test->rotationMatrix[4] = test->bounds->bbox[7][0];
		test->rotationMatrix[5] = test->bounds->bbox[7][2];
		test->rotationMatrix[6] = test->bounds->bbox[7][1];
		test->rotationMatrix[7] = 0;
		test->rotationMatrix[8] = test->bounds->bbox[6][0];
		test->rotationMatrix[9] = test->bounds->bbox[6][2];
		test->rotationMatrix[10] = test->bounds->bbox[6][1];
		test->rotationMatrix[11] = 0;

		obj[num].body = dBodyCreate (world);

		if ( test->objType == SPHERE ){
			dMassSetSphere (&mass,DENSITY,test->radius);
			obj[num].geom[0] = dCreateSphere (space,test->radius); // assumes one geometry per object
		}

		if ( test->objType == BOX ){
			dMassSetBox (&mass,DENSITY, lx, ly, lz);

			if (test->rootParent == test->item) // object is its own root parent
			{
				obj[num].geom[0] = dCreateBox (space, lx, ly, lz);
				test->odeNum = num;
			}
			else //child objects are first attached to geometry transforms
			{
				rp = FindObject( test->rootParent);
				numC = rp->numChildren;

				//create the next object
				obj[num].geom[0] = dCreateBox (space, lx, ly, lz);

				//add it to the child list of the root parent
				rp->cg[numC] = obj[num].geom[0],
				rp->numChildren++;
			}
		}

		if (test->objStatic == NONSTATIC )
		{
			if (test->parent == NULL)
			{
				dGeomSetBody(obj[num].geom[0],obj[num].body);
				dGeomSetPosition (obj[num].geom[0],test->xpos,test->zpos,test->ypos);
				dGeomSetRotation(obj[num].geom[0], test->rotationMatrix);
			}
			else
			{
				// Make a transpose rotation matrix from the parents rotation matrix
				KM_transpose34 ( invRotMatrix, rp->rotationMatrix );
				CalcAB(test->item, test->rootParent, objDelta);
				KM_dtransform ( objDelta, invRotMatrix, objDeltaRotated);

				test->objDelta[0] = objDeltaRotated[0];
				test->objDelta[1] = objDeltaRotated[1];
				test->objDelta[2] = objDeltaRotated[2];

				//Set the postion of the ODE geoms
				dGeomSetBody(obj[num].geom[0],obj[rp->odeNum].body);
				dGeomSetOffsetPosition (rp->cg[numC],objDeltaRotated[0],objDeltaRotated[2],objDeltaRotated[1]);
				dGeomSetRotation(rp->cg[numC], test->rotationMatrix);
			}
		}
		else //Object is static, don't attach it to a body
		{
			dGeomSetPosition (obj[num].geom[0],test->xpos,test->zpos,test->ypos);
			dGeomSetRotation(obj[num].geom[0], test->rotationMatrix);
		}

		//Will need to account for composite objects and their center of mass... later
		dBodySetMass (obj[num].body,&mass);

		// Clear old data from transform matrices
		if (test->m != NULL) {
			free(test->m);
			test->m = NULL;
		}

		// Create the new Matrix Array
		test->m = (LWDMatrix4*)malloc(totalFrames * sizeof(LWDMatrix4));

		test = test->next;
		num++;
	}


	////////////////////////////////////////////////
	// Start the simulation loop
	////////////////////////////////////////////////

	for (i=0;i<totalFrames;i++) {
		dSpaceCollide (space,0,&nearCallback);
		dWorldQuickStep (world,stepRate);

		if (DEBUG) {
			if (write_world) {
				FILE *f = fopen ("state.dif","wt");
				if (f) {
					dWorldExportDIF (world,f,"X");
					fclose (f);
				}
				write_world = 0;
			}
		}


		/////////////////////////////////////////////////////
		// Extract the geom motion from the sim for each step
		// Need to do this for each dGeomID (g)
		/////////////////////////////////////////////////////

		// reset num and test
		num = 0;
		test = objMatrixData;

		while (test != NULL)
		{
			rparent = FindObject( test->rootParent);
			pos = dGeomGetPosition (obj[num].geom[0]);
			R = dGeomGetRotation (obj[num].geom[0]);

			/////////////////////////////////////////////////////////
			// Generate the translation matrix for the current object
			// at the current frame.
			// Here we're converting an ODE matrix into something
			// Lightwave understands.
			/////////////////////////////////////////////////////////

			RM[0] = rparent->m[i][0][0];
			RM[8] = rparent->m[i][0][1];
			RM[4] = rparent->m[i][0][2];
			RM[3] = 0;
			RM[2] = rparent->m[i][1][0];
			RM[10] = rparent->m[i][1][1];
			RM[6] = rparent->m[i][1][2];
			RM[7] = 0;
			RM[1] = rparent->m[i][2][0];
			RM[9] = rparent->m[i][2][1];
			RM[5] = rparent->m[i][2][2];
			RM[11] = 0;

			test->m[i][0][0] = R[0];
			test->m[i][0][1] = R[8];
			test->m[i][0][2] = R[4];
			test->m[i][0][3] = 0;
			test->m[i][1][0] = R[2];
			test->m[i][1][1] = R[10];
			test->m[i][1][2] = R[6];
			test->m[i][1][3] = 0;
			test->m[i][2][0] = R[1];
			test->m[i][2][1] = R[9];
			test->m[i][2][2] = R[5];
			test->m[i][2][3] = 0;

			if (test->rootParent == test->item)
			{
				test->m[i][3][0] = pos[0] - test->xpos;
				test->m[i][3][1] = pos[2] - test->ypos;
				test->m[i][3][2] = pos[1] - test->zpos;
				test->m[i][3][3] = 1;
			}
			else
			{

				//Moves children perfectly but does not place them relative to parent
				test->m[i][3][0] = pos[0] - test->xpos;  
				test->m[i][3][1] = pos[2] - test->ypos;  
				test->m[i][3][2] = pos[1] - test->zpos;
				test->m[i][3][3] = 1;

			}

			num++;
			test = test->next;

		} // end while loop

		dJointGroupEmpty (contactgroup);

		sprintf (buf, "RefreshNow", NULL);
		local->evaluate( local->data, buf );

        sprintf (buf, "NextFrame", NULL);
		local->evaluate( local->data, buf );

	} //end for loop

	sprintf (buf, "GoToFrame %d", frameStart);
	local->evaluate( local->data, buf );

	dJointGroupDestroy (contactgroup);
	dSpaceDestroy (space);
	dWorldDestroy (world);

}


/*
======================================================================
nearCallback()

For ODE, taken from test_boxstack.cpp

This is called by dSpaceCollide when two objects in space are
potentially colliding.
====================================================================== */

static void nearCallback (void *data, dGeomID o1, dGeomID o2)
{
  int i;
  
  // Uncommenting this generates a few errors.  I don't know why the errors
  // occur or even what good the line does
  //if (o1->body && o2->body) return;

  /////////////////////////////////////////////////////////////////////////
  // exit without doing anything if the two bodies are connected by a joint
  /////////////////////////////////////////////////////////////////////////
  dBodyID b1 = dGeomGetBody(o1);
  dBodyID b2 = dGeomGetBody(o2);
  if (b1 && b2 && dAreConnectedExcluding (b1,b2,dJointTypeContact)) return;

  dContact contact[MAX_CONTACTS];   // up to MAX_CONTACTS contacts per box-box
  for (i=0; i<MAX_CONTACTS; i++) {
    contact[i].surface.mode = dContactBounce | dContactSoftCFM;
    contact[i].surface.mu = dInfinity;
    contact[i].surface.mu2 = 0;
    contact[i].surface.bounce = 0.1;
    contact[i].surface.bounce_vel = 0.1;
    contact[i].surface.soft_cfm = 0.01;
  }

  if (int numc = dCollide (o1,o2,MAX_CONTACTS,&contact[0].geom,sizeof(dContact)))
  {
	dMatrix3 RI;
    dRSetIdentity (RI);
	const dReal ss[3] = {0.02,0.02,0.02};
	for (i=0; i<numc; i++) {
		dJointID c = dJointCreateContact (world,contactgroup,contact+i);
		dJointAttach (c,b1,b2);
	}
  }
}


/*
======================================================================
KM_transpose4()
Calculates a transpose matrix.
This function could use some error checking.
====================================================================== */
void KM_transpose4 ( LWDMatrix4 n, LWDMatrix4* m, int i )
{

	//LWMAT_dinitm4( n, m[i][0][0],m[i][0][1],m[i][0][2],0,
	//				  m[i][1][0],m[i][1][1],m[i][1][2],0,
	//				  m[i][2][0],m[i][2][1],m[i][2][2],0,
	//			      0         ,0         ,0         ,1);
	LWMAT_dinitm4( n, m[i][0][0],m[i][1][0],m[i][2][0],0,
					  m[i][0][1],m[i][1][1],m[i][2][1],0,
					  m[i][0][2],m[i][1][2],m[i][2][2],0,
				      0         ,0         ,0         ,1);
}

/*
======================================================================
KM_transpose34()
Calculates a transpose matrix.
This function could use some error checking.
====================================================================== */
void KM_transpose34 ( dMatrix3 n, dMatrix3 m )
{
	KM_dinitm34( n, m[0],m[4],m[8],m[3],
					m[1],m[5],m[9],m[7],
					m[2],m[6],m[10],m[11]);
}

/*
======================================================================
KM_dinitm34()
Initiates a dMatrix3 matrix with the values provided.
====================================================================== */
void KM_dinitm34( dMatrix3 m, dReal a1, dReal b1, dReal c1, dReal d1,
							  dReal a2, dReal b2, dReal c2, dReal d2,
							  dReal a3, dReal b3, dReal c3, dReal d3 )
{
	m[0] = a1; m[1] = b1; m[2]  = c1; m[3]  = d1;
	m[4] = a2; m[5] = b2; m[6]  = c2; m[7]  = d2;
	m[8] = a3; m[9] = b3; m[10] = c3; m[11] = d3;
}


/*
======================================================================
KM_dmatmul34()
Multiplies two dMatrix3 matrices together resulting in a third matrix.

dMatrix3 a and dMatrix3 b are input
dMatrix3 c is the resultant matrix
====================================================================== */
void KM_dmatmul34( dMatrix3 a, dMatrix3 b, dMatrix3 c )
{
	c[0] = a[0]*b[0] + a[1]*b[4] + a[2]*b[8];
	c[1] = a[0]*b[1] + a[1]*b[5] + a[2]*b[9];
	c[2] = a[0]*b[2] + a[1]*b[6] + a[2]*b[10];
	c[3] = 0;
	c[4] = a[4]*b[0] + a[5]*b[4] + a[6]*b[8];
	c[5] = a[4]*b[1] + a[5]*b[5] + a[6]*b[9];
	c[6] = a[4]*b[2] + a[5]*b[6] + a[6]*b[10];
	c[7] = 0;
	c[8] = a[8]*b[0] + a[9]*b[4] + a[10]*b[8];
	c[9] = a[8]*b[1] + a[9]*b[5] + a[10]*b[9];
	c[10] = a[8]*b[2] + a[9]*b[6] + a[10]*b[10];
	c[11] = 0;
}


/*
======================================================================
KM_dtransform()
Converts a dMatrix3 to a LWDMatrix3, then multiplies a LWDVector by the
new LWDMatrix3 resulting in a rotated LWDVector.
====================================================================== */
void KM_dtransform( LWDVector a, dMatrix3 m, LWDVector b )
{
   int i;
   LWDMatrix3 n;

   LWMAT_dinitm3( n, m[0],m[8],m[4],
				     m[2],m[10],m[6],
					 m[1],m[9],m[5]);

   for ( i = 0; i < 3; i++ )
      b[ i ] = a[ 0 ] * n[ 0 ][ i ]
             + a[ 1 ] * n[ 1 ][ i ]
             + a[ 2 ] * n[ 2 ][ i ];
}


/*
======================================================================
CalcAB()
Calculates the AB vector between any two bodies. Rotation is not
accounted for.
This function could use some error checking.
====================================================================== */
int CalcAB (LWItemID id_a, LWItemID id_b, LWDVector delta )
{
	LWDVector objPosA;
	LWDVector objPosB;

	///////////////////////////////////////////////////////
	// Get the objects pivot position at the start frame 0
	///////////////////////////////////////////////////////
	iteminfo->param( id_a, LWIP_W_POSITION, 0, objPosA );
	iteminfo->param( id_b, LWIP_W_POSITION, 0, objPosB );

	///////////////////////////////////////
	// Get the relative position of B to A
	///////////////////////////////////////
	VSUB3(delta, objPosA, objPosB);

	return 0;
}


/*
======================================================================
CalcBB()
Calculate the bounding box and bounding sphere of an object.
This function could use some error checking.
====================================================================== */
int CalcBB ( LWItemID item )
{
	LWMeshInfoID mesh;
	ObjectData* test;
	int i,j;

	// get the current object
	test = FindObject(item);

	// initialize the first to vectors in the array
	for (i=0;i<2;i++)
		for (j=0;j<3;j++)
			test->bounds->bbox[i][j] = 0.0;

	// initialize the radius
	test->radius = 0.0;

	test->bounds->id = item; // should delete from bbox

	iteminfo->param( item, LWIP_RIGHT, 0, test->bounds->bbox[2]);
	iteminfo->param( item, LWIP_UP, 0, test->bounds->bbox[3]);
	iteminfo->param( item, LWIP_FORWARD, 0, test->bounds->bbox[4]);
	iteminfo->param( item, LWIP_W_RIGHT, 0, test->bounds->bbox[5]);
	iteminfo->param( item, LWIP_W_UP, 0, test->bounds->bbox[6]);
	iteminfo->param( item, LWIP_W_FORWARD, 0, test->bounds->bbox[7]);
	iteminfo->param( item, LWIP_W_POSITION, 0, test->bounds->bbox[8]);
	iteminfo->param( item, LWIP_PIVOT, 0, test->bounds->bbox[9]);
	iteminfo->param( item, LWIP_POSITION, 0, test->bounds->bbox[10]);

	// get the mesh info
	mesh = objinfo->meshInfo(item, 1);

	// scan all points
	mesh->scanPoints( mesh, (LWPntScanFunc*)pntScanBB, test );

	return 0;

}

/*
======================================================================
pntScanBB()

Point scan callback for Bounding Boxes and Radius.  Bounds is an array of two
LWDVectors.  It will contain two points which can define the entire
bounding box.  Radius is a float.
====================================================================== */
static int pntScanBB( ObjectData* test, LWPntID id )
{
	LWMeshInfo *mesh;
	LWFVector p,q;
	float radius;
	float x,y,z;
	int i;

	// get the point position
	mesh = objinfo->meshInfo(test->item, 1);
	mesh->pntOtherPos(mesh, id, q);

	// transform to item coordinates
	for ( i = 0; i < 3; i++ )
		p[ i ] = ( q[0] - test->bounds->bbox[8][0] ) * test->bounds->bbox[5][i]
			   + ( q[1] - test->bounds->bbox[8][1] ) * test->bounds->bbox[6][i]
			   + ( q[2] - test->bounds->bbox[8][2] ) * test->bounds->bbox[7][i]
			   + test->bounds->bbox[9][ i ];

	// test for min/max bound
	for ( i = 0; i < 3; i++ )
	{
		if (p[i] < test->bounds->bbox[0][i])
			test->bounds->bbox[0][i] = p[i];

		if (p[i] > test->bounds->bbox[1][i])
			test->bounds->bbox[1][i] = p[i];
	}

	x = p[0];
	y = p[1];
	z = p[2];


	radius = sqrt((x*x) + (y*y) + (z*z));

	// find the largest radius. Not nescessarily the bounds
	if (test->radius < radius) test->radius = radius;

   return 0;
}


/*
======================================================================
findRootParents()

Recursive function to find the root parent of each object in the
objMatrixData Linked List.  Not interested in hierarchy of
parent/child relationships, just which objects are the 'root' objects
and which are children, independent of how many levels down.
Just being a rootParent doesn't necessarily mean the object has
anything parented to it.
====================================================================== */

LWItemID findRootParents(ObjectData* test)
{
	LWItemID rootP = NULL;

	//////////////////////////////////////////////////////////////////
	// Check each object for parent data
	// This can be simplified nicely later
	//////////////////////////////////////////////////////////////////

	if (test->parent == NULL) return test->item; //a root parent is found

	if (test->rootParent != NULL) // node already tested
	{
		return test->rootParent;
	}
	else
	{
		rootP = findRootParents(FindObject(test->parent));
	}

	return rootP;
}


/*
======================================================================
sortList()

Sorting function to place root parents at the front of the ObjectData
linked list.

The idea is simple, iterate over the list once. As root parents are
found, move them to the front of the list.

No error checking implemented.
====================================================================== */

int sortList()
{

	ObjectData* test = objMatrixData; //start at the second object
	ObjectData* next = NULL;
	ObjectData* last = NULL;

	if (test == NULL || test->next == NULL) //one object
	{
		return 0;
	}

	while (test != NULL) //two or more
	{
		if (test == objMatrixData)//skip first object
		{
			last = test;
			test = test->next;
			continue;
		}

		if (test->parent == NULL) //test for root parent
		{
			next = test->next;
			last->next = test->next;
			test->next = objMatrixData;
			objMatrixData = test;		
		}
		else
		{
			last = test;
		}

		test = last->next;
	}

	return 0;
}


/*
======================================================================
Start Displacement here
Base code from Inertia.c, thanks Ernie and Newtek
======================================================================

======================================================================
popCnt_VMAP()

Return the number of weight maps, plus 1 for "(none)".  An xpanel
callback for the vmap popup list, also used by Load().
====================================================================== */

XCALL_( static int )
popCnt_VMAP( void *data )
{
   return 1 + objf->numVMaps( LWVMAP_WGHT );
}


/*
======================================================================
popName_VMAP()

Return the name of a vmap, given an index.  An xpanel callback for the
vmap popup list, also used by Load() and Save().
====================================================================== */

XCALL_( static const char * )
popName_VMAP( void *data, int idx )
{
   if ( idx == 0 ) return "(none)";

   return objf->vmapName( LWVMAP_WGHT, idx - 1 );
}


/*
======================================================================
popCnt_Item()

Return the number of objects, plus 1 for "(none)".  An xpanel
callback for the pivot object popup list.
====================================================================== */

XCALL_( static int )
popCnt_Item( void *data )
{
   int n = 1;
   LWItemID id;

   id = iteminfo->first( LWI_OBJECT, NULL );
   while ( id ) {
      n++;
      id = iteminfo->next( id );
   }

   return n;
}


/*
======================================================================
popName_Item()

Return the name of an object, given an index.  An xpanel callback for
the pivot object popup list.
====================================================================== */

XCALL_( static const char * )
popName_Item( void *data, int idx )
{
   const char *a;
   int i = 0;
   LWItemID id;

   if ( idx == 0 ) return "(none)";

   idx--;

   if ( id = iteminfo->first( LWI_OBJECT, NULL ) )
    a = iteminfo->name( id );

   while ( id && i < idx ) {
      i++;
      if ( id = iteminfo->next( id ))
         a = iteminfo->name( id );
   }

   return a;
}


/*
======================================================================
popItemID()

Return the item ID of an object, given an index.
====================================================================== */

static LWItemID popItemID( int idx )
{
   LWItemID id;
   int i = 0;

   if ( idx <= 0 ) return NULL;
   idx--;

   id = iteminfo->first( LWI_OBJECT, NULL );
   while ( id && i < idx ) {
      i++;
      id = iteminfo->next( id );
   }

   return id;
}


/*
======================================================================
AddObject()

Add an object to the end of the ODE Linked List.

This is where we would determine if an object had a parent.
====================================================================== */

XCALL_( static LWError )
AddObject(LWItemID id, ODEDisplaceData *dat)
{

	ObjectData* test = objMatrixData;
	LWDVector objPos;
	LWItemID parentID;


	//////////////////////////////////
	// Find the end of the LinkedList
	//////////////////////////////////

	while (test != NULL && test->next != NULL)
	{
		test=test->next;
	}

	///////////////////////////////////////////////////////
	// Get the objects pivot position at the start frame 0
	///////////////////////////////////////////////////////
	iteminfo->param( dat->item, LWIP_W_POSITION, 0, objPos );

	///////////////////////////////////////////////////////
	// Get the parent object, if any, otherwise null
	///////////////////////////////////////////////////////
	parentID = iteminfo->parent(id);

	//////////////////////////////////////////////////////////////
	// Create and add the new object to the end of the linked list
	//////////////////////////////////////////////////////////////
	
	if (objMatrixData == NULL) {
		objMatrixData = (ObjectData*)malloc(sizeof(ObjectData));
		test = objMatrixData;
	}
	else {
		test->next = (ObjectData*)malloc(sizeof(ObjectData));
		test=test->next;
	}

	// initialize bounds
	test->bounds = (BBox*)malloc(sizeof(BBox));

	test->m = NULL;
	test->item = id;
	test->parent = parentID;
	test->rootParent = NULL;
	test->numChildren = 0;
	test->odeNum = 0;
	test->next = NULL;
	test->objType = SPHERE;  // default
	test->objStatic = NONSTATIC;
	test->xpos = objPos[0];
	test->ypos = objPos[1];
	test->zpos = objPos[2];

	return AFUNC_OK;
}


/*
======================================================================
RemoveObject()

Remove an object from the ODE Linked List
====================================================================== */

void RemoveObject(LWItemID id)
{

	ObjectData* test = objMatrixData;
	ObjectData* last = objMatrixData;

	//////////////////////////////////////////////////
	// Trying to remove a Linked node when none exist
	//////////////////////////////////////////////////
	if (objMatrixData == NULL)
		return;

	// Only one object
	if (objMatrixData->item == id && objMatrixData->next == NULL) {

		free (objMatrixData->m);
		objMatrixData->m = NULL;

		free (objMatrixData->bounds);
		objMatrixData->bounds = NULL;

		free (objMatrixData);
		objMatrixData = NULL;

		return;
	}

	// Only two objects
	if (objMatrixData->item == id) {

		test = objMatrixData->next;

		free (objMatrixData->m);
		objMatrixData->m = NULL;

		free (objMatrixData->bounds);
		objMatrixData->bounds = NULL;

		free (objMatrixData);
		objMatrixData = test;

		return;
	}

	// Find the object 
	while (test->item != id)
	{
		last = test;
		test = test->next;
	}

	// Determine if it is last or not
	if (test->next != NULL){
		last->next = test->next;
	}
	else {
		last->next = NULL;
	}

	free (test->m);
	test->m = NULL;

	free (test->bounds);
	test->bounds = NULL;

	free (test);
	test = NULL;

}


/*
======================================================================
ObjectData* FindObject()

Find an object in the ODE Linked List
====================================================================== */

ObjectData* FindObject( LWItemID id)
{

	ObjectData* test = objMatrixData;

	/////////////////////
	// Empty Linked List
	/////////////////////
	if (test == NULL) return NULL;


	//////////////////////////////////////////////////
	// Find the object with the requested LWItemID
	//////////////////////////////////////////////////
	while (test->item != id)
	{
		test = test->next;
		if (test == NULL) return NULL;
	}

	return test;
}


/*
======================================================================
Create()

Handler callback.  Allocate and initialize instance data.
====================================================================== */

XCALL_( static LWInstance )
Create( void *priv, LWItemID item, LWError *err )
{
   ODEDisplaceData *dat;

   if ( dat = (struct st_ODEDisplaceData *)calloc( 1, sizeof( ODEDisplaceData ))) {
      dat->item = item;
   }

   ////////////////////////////////////////////////
   // Add the object to the ODE object linked list
   ////////////////////////////////////////////////
   AddObject(item, dat);

   return dat;
}


/*
======================================================================
Destroy()

Handler callback.  Free resources allocated by Create().
====================================================================== */

XCALL_( static void )
Destroy( ODEDisplaceData *dat )
{

	/////////////////////////////////////////////////////
	// Clear an object from the ODE object linked list
	/////////////////////////////////////////////////////

	RemoveObject(dat->item);
	
	if( dat ) {
		free( dat );
		dat = NULL;
	}
}


/*
======================================================================
Copy()

Handler callback.  Copy instance data.  If your instance data contains
allocated resources, note that a simple *to = *from is insufficient.
====================================================================== */

XCALL_( static LWError )
Copy( ODEDisplaceData *to, ODEDisplaceData *from )
{
   LWItemID id;

   id = to->item;
   *to = *from;
   to->item = id;

   return AFUNC_OK;
}


/*
======================================================================
Load()

Handler callback.  Read instance data.
====================================================================== */

XCALL_( static LWError )
Load( ODEDisplaceData *dat, const LWLoadState *ls )
{
   ObjectData* item;
   short ver;

   item = FindObject( dat->item );

   LWLOAD_I2( ls, &ver, 1 );
   LWLOAD_I4( ls, (long*)&dat->objType, 1 );
   LWLOAD_I4( ls, (long*)&dat->objStatic, 1 );
   
   // Need to fix objType to be consistant
   item->objType = dat->objType + 1;
   item->objStatic = dat->objStatic;

   return AFUNC_OK;
}


/*
======================================================================
Save()

Handler callback.  Write instance data.
====================================================================== */

XCALL_( static LWError )
Save( ODEDisplaceData *dat, const LWSaveState *ss )
{
   short ver = ODEDISPLACE_VERSION;

   LWSAVE_I2( ss, &ver, 1 );
   LWSAVE_I4( ss, (long*)&dat->objType, 1 );
   LWSAVE_I4( ss, (long*)&dat->objStatic, 1 );
	
   return AFUNC_OK;
}


/*
======================================================================

Describe()

Handler callback.  Write a short, human-readable string describing
the instance data.
====================================================================== */

XCALL_( static const char * )
Describe( ODEDisplaceData *dat )
{
	sprintf( dat->desc, " ODE v0.2.0 BETA", NULL );
   return dat->desc;
}


/*
======================================================================
UseItems()

Handler callback.  Return an array of items we depend on.
====================================================================== */

XCALL_( static const LWItemID * )
UseItems( ODEDisplaceData *dat )
{
	return NULL;
}


/*
======================================================================
ChangeID()

Handler callback.  ID numbers for items in the scene aren't constant.
They can change when items are added or deleted.  If the ID of our
pivot object has changed, we update our instance data.
====================================================================== */

XCALL_( static void )
ChangeID( ODEDisplaceData *dat, const LWItemID *ids )
{
   //int i = 0;

   //while ( ids[ i ] )
   //   if ( ids[ i ] == dat->pivot ) {
   //      dat->pivot = ids[ i + 1 ];
   //      return;
   //   }
   //   else
   //      i += 2;

   return;
}


/*
======================================================================
Init()

Handler callback, called at the start of rendering.
====================================================================== */

XCALL_( static LWError )
Init( ODEDisplaceData *dat, int mode )
{
   return NULL;
}


/*
======================================================================
Cleanup()

Handler callback, called at the end of rendering.
====================================================================== */

XCALL_( static void )
Cleanup( ODEDisplaceData *dat )
{
   return;
}


/*
======================================================================
NewTime()

Handler callback, called at the start of each sampling pass.
====================================================================== */

XCALL_( static LWError )
NewTime( ODEDisplaceData *dat, LWFrame fr, LWTime t )
{
   dat->frame = fr;
   dat->time = t;
   //dat->vmId = NULL;  // reset now, re-aquire on first vertex of this pass

   /* The following operations can be performed once every time,
      rather than once per vertex. */

   /* get the rotation origin for the pivot object */
   //if ( dat->pivot )
   //   iteminfo->param( dat->pivot,
   //      ( dat->flags & INF_PIVLOCAL ) ? LWIP_POSITION : LWIP_W_POSITION,
   //      dat->time, dat->pivpos );
   //else
   //   iteminfo->param( dat->item, LWIP_PIVOT, dat->time, dat->pivpos );

   /* get the rate from the envelope */
   //vparmf->getVal( dat->lagParm, dat->time, NULL, &dat->lagRate );

   return NULL;
}


/*
======================================================================
Flags()

Handler callback.
====================================================================== */

XCALL_( static int )
Flags( ODEDisplaceData *dat )
{
   //return( dat->flags & INF_PIVLOCAL ) ? 0 : LWDMF_WORLD;
   return 0;
}


/*
======================================================================
Evaluate()

Handler callback.  This is called for each vertex of the object the
instance is associated with.  The vertex is moved by setting the
members of the source[] array in the displacement access structure.
====================================================================== */

XCALL_( static void )
Evaluate( ODEDisplaceData *dat, LWDisplacementAccess *da )
{
	LWDMatrix4 objRot = {0.0};
	LWDMatrix4 objRotT = {0.0};
	LWDVector vertexPos;
	ObjectData* test;
	int cf;  // Current Frame - are fractional frames a concern here?


	////////////////////////////////////////////////////////////////////////
	// All that needs to be done is to get the correct transfomation matrix
	// array for the current object and multiply it against the vertices
	////////////////////////////////////////////////////////////////////////

	// Get the current frame
	cf = dat->frame;

	// Get the current object structure
	test = FindObject(dat->item);

	// Quit the eval function if this is a newly minted object
	// with no ODE transform data
	if (test == NULL || test->m == NULL) return;
	
	// Matrix to rotate from item to world coordinates
	objRot[0][0] = test->bounds->bbox[5][0];
	objRot[0][1] = test->bounds->bbox[5][1];
	objRot[0][2] = test->bounds->bbox[5][2];
	objRot[0][3] = 0;
	objRot[1][0] = test->bounds->bbox[6][0];
	objRot[1][1] = test->bounds->bbox[6][1];
	objRot[1][2] = test->bounds->bbox[6][2];
	objRot[1][3] = 0;
	objRot[2][0] = test->bounds->bbox[7][0];
	objRot[2][1] = test->bounds->bbox[7][1];
	objRot[2][2] = test->bounds->bbox[7][2];
	objRot[2][3] = 0;
	objRot[3][0] = 0;
	objRot[3][1] = 0;
	objRot[3][2] = 0;
	objRot[3][3] = 1;

	// Put the coordinates for the current vertex into a LWDVector
	vertexPos[0] = da->source[ 0 ];
	vertexPos[1] = da->source[ 1 ];
	vertexPos[2] = da->source[ 2 ];

	// Multiply the coordinate by the matrix for the correct object at the current frame
	LWMAT_dmatmul4 (test->m[cf], objRot, objRotT); //transforms to the objects local coord frame
	LWMAT_dtransformp( vertexPos, objRotT, da->source);

	// Done tranformation
}

/*
======================================================================
ODEDisplace()

Handler activation function.  Check the version, get some globals, and
fill in the callback fields of the handler structure.
====================================================================== */

XCALL_( int )
ODEDisplace( long version, GlobalFunc *global, LWDisplacementHandler *local, void *serverData)
{
   if ( version != LWDISPLACEMENT_VERSION )
      return AFUNC_BADVERSION;

   objf     = (struct st_LWObjectFuncs *)global( LWOBJECTFUNCS_GLOBAL,   GFUSE_TRANSIENT );
   envf     = (struct st_LWEnvelopeFuncs *)global( LWENVELOPEFUNCS_GLOBAL, GFUSE_TRANSIENT );
   chanf    = (struct st_LWChannelInfo *)global( LWCHANNELINFO_GLOBAL,   GFUSE_TRANSIENT );
   iteminfo = (struct st_LWItemInfo *)global( LWITEMINFO_GLOBAL,      GFUSE_TRANSIENT);


   if ( !objf || !envf || !chanf || !iteminfo )
      return AFUNC_BADGLOBAL;

   local->inst->create   = Create;
   local->inst->destroy  = (void (*)(void *))Destroy;
   local->inst->load     = (const char *(*)(void *,const struct st_LWLoadState *))Load;
   local->inst->save     = (const char *(*)(void *,const struct st_LWSaveState *))Save;
   local->inst->copy     = (const char *(__cdecl *)(void *,void *))Copy;
   local->inst->descln   = (const char *(__cdecl *)(void *))Describe;

   local->item->useItems = (void *const *(__cdecl *)(void *))UseItems;
   local->item->changeID = (void (__cdecl *)(void *,void *const * ))ChangeID;

   local->rend->init     = (const char *(__cdecl *)(void *,int))Init;
   local->rend->cleanup  = (void (__cdecl *)(void *))Cleanup;
   local->rend->newTime  = (const char *(__cdecl *)(void *,int,double))NewTime;

   local->evaluate       = (void (__cdecl *)(void *,struct st_LWDisplacementAccess *))Evaluate;
   local->flags          = (unsigned int (__cdecl *)(void *))Flags;

   return AFUNC_OK;
}


/* ----- all of this stuff is for the interface ----- */

/*
======================================================================
ODEDisplaceData_get()

XPanel callback.  XPanels calls this when it wants to retrieve the
value of a control from your instance data.  Returns a pointer to the
field containing the requested value.
====================================================================== */
XCALL_( static void * )
ODEDisplaceData_get( ODEDisplaceData *dat, unsigned long vid )
{
   void *result = NULL;

   if ( dat )
      switch ( vid ) {
         case CH_TYPE:
            result = &dat->objType;
            break;
         case CH_STATIC:
            result = &dat->objStatic;
            break;
      }

   return result;
}

/*
======================================================================
ODEDisplaceData_set()

XPanel callback.  XPanels calls this when it wants to store the value
of a control in your instance data.  Returns 1 if the value could be
stored, otherwise 0.
====================================================================== */
XCALL_( static int )
ODEDisplaceData_set( ODEDisplaceData *dat, unsigned long vid, void *value )
{
   int rc = 0;
   ObjectData* object;

   object = FindObject(dat->item);

   if ( dat )
      switch ( vid ) {
         case CH_TYPE:
			 dat->objType = *(( int * ) value );
			 	if (dat->objType == 0)
					object->objType = SPHERE;
				else if (dat->objType == 1)
					object->objType = BOX;
			 rc = 1;
			 break;

         case CH_STATIC:
			 dat->objStatic = *(( int * ) value );
			 	if (dat->objStatic == 0)
					object->objStatic = NONSTATIC;
				else if (dat->objStatic == 1)
					object->objStatic = STATIC;
			 rc = 1;
			 break;
      }

   return rc;
}

/*
======================================================================
ChangeNotify()

Xpanel callback.  XPanels calls this when an event occurs that affects
the value of one of your controls.  We use the instance update global
to tell Layout that our instance data has changed.
====================================================================== */
XCALL_( static void )
ChangeNotify( LWXPanelID panID, unsigned long cid, unsigned long vid, int event_code )
{
   void *val = NULL;
   void *dat;

   if ( event_code == LWXPEVENT_VALUE )
      if ( dat = xpanf->getData( panID, 0 ))
         lwupdate( LWDISPLACEMENT_HCLASS, dat );
}


/*
======================================================================
get_xpanel()

Create and initialize an xpanel.
====================================================================== */

static LWXPanelID get_xpanel( GlobalFunc *global, ODEDisplaceData *dat )
{
   LWXPanelID panID = NULL;

   static LWXPanelHint hint[] = {
      XpLABEL( 0, "ODEDisplace" ),
      XpCHGNOTIFY( ChangeNotify ),
	  XpSTRLIST( CH_TYPE, objType_strlist),
      XpENABLE_( CH_TYPE ), XpEND,
      XpEND
   };

   xpanf = (struct st_LWXPanelFuncs *)global( LWXPANELFUNCS_GLOBAL, GFUSE_TRANSIENT );
   if ( xpanf ) {
      panID = xpanf->create( LWXP_VIEW, ctrl_list );
      if ( panID ) {
         xpanf->hint( panID, 0, hint );
         xpanf->describe(
			 panID,
			 data_descrip,
			 (void *(__cdecl *)(void *,unsigned long))ODEDisplaceData_get,
			 (enum en_LWXPRefreshCodes (__cdecl *)(void *,unsigned long,void *))ODEDisplaceData_set);
         xpanf->viewInst( panID, (void *(__cdecl *)(void *,unsigned long))dat );
         xpanf->setData( panID, 0, dat );
      }
   }

   return panID;
}


/*
======================================================================
ODEDisplace_UI()

Interface activation function.  Get a global, create an xpanel, and
fill in the fields of the LWInterface structure.
====================================================================== */

XCALL_( int )
ODEDisplace_UI( long version, GlobalFunc *global, LWInterface *local, void *serverData )
{
   if ( version != LWINTERFACE_VERSION )
      return AFUNC_BADVERSION;

   lwupdate = (void (__cdecl *)(const char *,void *))global( LWINSTUPDATE_GLOBAL, GFUSE_TRANSIENT );
   if ( !lwupdate )
      return AFUNC_BADGLOBAL;

   local->panel = get_xpanel( global, (struct st_ODEDisplaceData *)local->inst );
   if ( !local->panel )
      return AFUNC_BADGLOBAL;

   local->options = NULL;
   local->command = NULL;

   return AFUNC_OK;
}


/*
======================================================================
This is the server description.  LightWave looks at this first to
determine what plug-ins the file contains.  It lists each plug-in's
class and internal name, along with a pointer to the activation
function.  You can optionally add a user name, or more than one in
different languages, if you like.
====================================================================== */
extern "C" {

ServerRecord ServerDesc[] = {
   { LWLAYOUTGENERIC_CLASS, "ODE_Generic",
	(int (*)(long,void * (*)(const char *,int),void *,void *))ODEGeneric },
   { LWDISPLACEMENT_HCLASS, "ODE_Displace",
	(int (*)(long,void * (*)(const char *,int),void *,void *))ODEDisplace },
   { LWDISPLACEMENT_ICLASS, "ODE_Displace",
	(int (*)(long,void * (*)(const char *,int),void *,void *))ODEDisplace_UI },
   { NULL }
};

}
